library geiger_localstorage;

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:geiger_localstorage/src/db/data/switchable_boolean.dart';
import 'package:geiger_localstorage/src/serialization/byte_sink.dart';
import 'package:geiger_localstorage/src/serialization/byte_stream.dart';
import 'package:geiger_localstorage/src/serialization/serializer_helper.dart';

/// <p>The implementation of the node interface.</p>
///
/// <p>This Class denotes one node containing sub-nodes in a tree-like structure. Each node may have
/// n children. Each node may be a skeleton-only node (contains only name and a reference to a
/// mapper), or may be materialized (contains all data). Typically when fetching a node, the node
/// is materialized but its sub-nodes are skeleton-only nodes. All skeleton nodes materialize
/// automatically if their data is accessed.</p>
class NodeImpl implements Node {
  static const bool _debug = true;
  static const List<Field> _skeletonOrdinals = <Field>[
    Field.path,
    Field.name,
    Field.tombstone
  ];

  static const int serialversionUID = 11239348938;

  /// an indicator whether the current object is a skeleton
  final SwitchableBoolean _skeleton = SwitchableBoolean(false);

  /// Contains the mapper for a skeleton to fetch any subsequent  data
  StorageController? _controller;

  /// contains the ordinals of a node
  final Map<Field, String> _ordinals = <Field, String>{};

  /// contains the key/value pairs of a node
  final Map<String, NodeValue> _values = {};

  /// Holds all child nodes as tuples, where the name is used as a key and
  /// the value is of type StorageNode
  final Map<String, Node> _childNodes = {};

  /// Constructor creating a skeleton node.
  ///
  /// @param path       the path of the node
  /// @param controller the controller to fetch the full node
  NodeImpl.createSkeleton(String path, String nodeOwner, this._controller) {
    _skeleton.set(true);
    owner = nodeOwner;
    try {
      set(Field.path, path);
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  static Future<Node> fromNode(Node node) async {
    Node ret = NodeImpl(node.name, node.owner);
    await ret.update(node);
    return ret;
  }

  /// Create a empty node for the given path.
  ///
  /// @param path the node path
  /// @param isTombstone true if the node is a tombstone node
  /// @param visibility the visibility of the node or null if default
  /// @param nodeValues the node values to be stored or null if none
  /// @param childNodes the child nodes to be included or null if none
  static Future<Node> fromPath(String path, String owner,
      {bool? isTombstone,
      Visibility? visibility,
      List<NodeValue>? nodeValues,
      List<Node>? childNodes}) async {
    var node = NodeImpl(getNameFromPath(path)!, owner, getParentFromPath(path));
    try {
      if (isTombstone != null) {
        node.set(Field.tombstone, isTombstone ? 'true' : 'false');
      }
      if (visibility != null) {
        node.set(Field.visibility, visibility.toValueString());
      }
      if (nodeValues != null) {
        for (var nv in nodeValues) {
          await node.addValue(nv);
        }
      }
      if (childNodes != null) {
        for (var n in childNodes) {
          await node.addChild(n);
        }
      }
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
    return node;
  }

  /// <p>creates a fully fledged new empty node.</p>
  ///
  /// @param name   the name for the node
  /// @param parent the parent of the node (may be null if root node is the parent)
  /// @param vis    visibility of the node
  NodeImpl(String name, String nodeOwner,
      [String? parent, Visibility vis = Visibility.red]) {
    owner = nodeOwner;
    if (parent == null) {
      parent = getParentFromPath(name);
      name = getNameFromPath(name) ?? '';
    }
    _ordinals[Field.path] = '$parent${GenericController.pathDelimiter}$name';
    _ordinals[Field.visibility] = vis.toString();
    _skeleton.set(false);
  }

  /// <p>Converts current node into a materialized node from a skeleton.</p>
  Future<void> _init() async {
    if (skeleton) {
      await update(await controller!.get(path));
      _skeleton.set(false);
      _controller = null;
    }
  }

  /// <p>Returns the name part from a fully qualified node path.</p>
  ///
  /// @param path the fully qualified path of which the name is extracted
  /// @return the name part of the path
  static String? getNameFromPath(String? path) {
    if (path == null) {
      return null;
    }
    return path
        .substring(path.lastIndexOf(GenericController.pathDelimiter) + 1);
  }

  /// <p>Returns the fully qualified node path of the parental node.</p>
  ///
  /// @param path the fully qualified path of which the parental node is extracted
  /// @return the fully qualified path to the parental node
  static String? getParentFromPath(String? path) {
    if (path == null) {
      return null;
    }
    if (!path.contains(GenericController.pathDelimiter)) {
      return '';
    }
    return path.substring(0, path.lastIndexOf(GenericController.pathDelimiter));
  }

  @override
  Future<NodeValue?> getValue(String key) async {
    await _init();
    NodeValue? ret = _values[key];
    if (ret != null) {
      ret = ret.deepClone();
    }
    return ret;
  }

  @override
  Future<NodeValue> updateValue(NodeValue value) async {
    await _init();
    NodeValue? ret = await getValue(value.key);
    if (ret == null) {
      throw StorageException(
          'Value ' + value.key + ' not found in node ' + (name));
    }
    _values[value.key] = value;

    return ret;
  }

  @override
  Future<void> addValue(NodeValue value) async {
    await _init();
    if (await getValue(value.key) != null) {
      throw StorageException(
          'value "${value.key}" does already exist in the node "$path"');
    }
    _values[value.key] = value;
  }

  @override
  Future<void> addOrUpdateValue(NodeValue value) async {
    _values[value.key] = value;
  }

  @override
  Future<NodeValue?> removeValue(String key) async {
    await _init();
    return _values.remove(key);
  }

  @override
  Future<void> addChild(Node node) async {
    await _init();
    String key = node.name;
    if (!_childNodes.containsKey(key)) {
      _childNodes[key] = node;
    }
  }

  @override
  String get owner {
    try {
      return _ordinals[Field.owner] ?? '<UNSET>';
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  set owner(String newOwner) {
    if (newOwner.contains('\\:')) {
      throw StorageException('owner may not contain ":" character');
    }
    try {
      _ordinals[Field.owner] = newOwner;
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  String get name {
    try {
      return _getSkeleton(Field.name) ?? '';
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  String? get parentPath {
    try {
      return getParentFromPath(_getSkeleton(Field.path));
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  String get path {
    try {
      return _getSkeleton(Field.path) ?? '';
    } on StorageException {
      throw Exception('Oops.... this should not happen... contact developer');
    }
  }

  @override
  Visibility get visibility {
    var rawVisibility = _ordinals[Field.visibility];
    return VisibilityExtension.valueOf(rawVisibility!) ?? Visibility.red;
  }

  @override
  set visibility(Visibility newVisibility) {
    _ordinals[Field.visibility] = newVisibility.toString();
  }

  @override
  Future<Map<String, NodeValue>> getValues() async {
    return Map.from(_values);
  }

  String? _getSkeleton(Field field) {
    if (field == Field.name) {
      return getNameFromPath(_ordinals[Field.path]);
    } else if (!_skeletonOrdinals.contains(field)) {
      throw Exception('Illegal access to non skeletal ordinal');
    } else {
      return _ordinals[field];
    }
  }

  /// <p>Gets an ordinal field of the node.</p>
  /// @param field the ordinal field to get
  /// @return the currently set value
  /// @throws StorageException if a field does not exist
  Future<String?> get(Field field) async {
    if (!_skeletonOrdinals.contains(field)) {
      await _init();
      switch (field) {
        case Field.owner:
        case Field.path:
        case Field.visibility:
        case Field.lastModified:
        case Field.expiry:
          return _ordinals[field];
        case Field.name:
          return getNameFromPath(_ordinals[Field.path]);
        default:
          throw StorageException('unable to fetch field ' + field.toString());
      }
    } else {
      return _getSkeleton(field);
    }
  }

  /// <p>Sets an ordinal field of the node.</p>
  ///
  /// @param field the ordinal field to be set
  /// @param value the new value
  /// @return the previously set value
  /// @throws StorageException if a field does not exist
  Future<String?> set(Field field, String? value) async {
    if (!_skeletonOrdinals.contains(field)) {
      await _init();
    }

    var current = _ordinals[field];

    if (field != Field.lastModified &&
        ((current != null && current != value) ||
            (current == null && value != null))) {
      touch();
    }
    switch (field) {
      case Field.owner:
      case Field.path:
      case Field.visibility:
      case Field.lastModified:
      case Field.expiry:
        if (value == null) {
          return _ordinals.remove(field);
        }
        var prev = _ordinals[field];
        _ordinals[field] = value;
        return prev;
      case Field.tombstone:
        var prev = _ordinals[field] ?? 'false';
        _ordinals[field] = ('true' == value) ? 'true' : 'false';
        return prev;
      default:
        throw StorageException('unable to set field ' + field.toString());
    }
  }

  /// <p>Checks if the current node is marked as TOMBSTONE (old deleted node).</p>
  /// @return true if node is a tombstone
  @override
  bool get tombstone {
    try {
      return 'true' == _getSkeleton(Field.tombstone);
    } on StorageException {
      throw Exception('OOPS! Unexpected exception... please contact developer');
    }
  }

  @override
  Future<void> removeChild(String name) async {
    _childNodes.remove(name);
  }

  @override
  Future<Map<String, Node>> getChildren() async {
    await _init();
    Map<String, Node> ret = {};
    for (MapEntry<String, Node> entry in _childNodes.entries) {
      ret[entry.key] = await entry.value.deepClone();
    }
    return ret;
  }

  @override
  Future<Node?> getChild(String name) async {
    await _init();
    return _childNodes[name];
  }

  @override
  @Deprecated('This call will be removed in the final release')
  Future<String> getChildNodesCsv() async {
    await _init();
    if (_childNodes.isEmpty) {
      return '';
    }
    var csv = '';
    for (var s in _childNodes.keys) {
      csv += s + ',';
    }
    return csv.substring(0, csv.length - 1);
  }

  @override
  bool get skeleton {
    return _skeleton.get();
  }

  @override
  StorageController? get controller {
    return _controller;
  }

  @override
  set controller(StorageController? controller) {
    _controller = controller;
  }

  @override
  bool equals(Object? n2) {
    if (n2 == null || n2 is! NodeImpl) {
      if (_debug) {
        print('## Equals: #1');
      }
      return false;
    }
    // check if both are in same skeletonized state
    if (skeleton != n2.skeleton) {
      return false;
    }

    if (!skeleton) {
      // check ordinals
      if (_ordinals.length != n2._ordinals.length) {
        if (_debug) {
          print('## Equals: #5');
        }
        return false;
      }
      for (Field e in Field.values) {
        try {
          if (n2._ordinals[e] != _ordinals[e]) {
            if (_debug) {
              print('## Equals: #6');
            }
            return false;
          }
        } on Exception {
          if (_debug) {
            print('## Equals: #7');
          }
          throw Exception('Oops.... this should not happen... contact developer');
        }
      }

      //check values
      if (_values.length != n2._values.length) {
        if (_debug) {
          print('## Equals: #8');
        }
        return false;
      }
      for (var e in _values.entries) {
        try {
          if (!(e.value.equals(n2.getValue(e.key)))) {
            if (_debug) {
              print('## Equals: #9');
            }
            return false;
          }
        } on StorageException {
          if (_debug) {
            print('## Equals: #10');
          }
          return false;
        }
      }

      //check child nodes
      if (_childNodes.length != n2._childNodes.length) {
        if (_debug) {
          print(
              '## Equals: #11 difference in child nodes (local: ${_childNodes.length}; remote:${n2._childNodes.length})');
        }
        return false;
      }
      for (var n in _childNodes.keys) {
        if (n2._childNodes[n] == null) {
          if (_debug) {
            print('## Equals: #12');
          }
          return false;
        }
      }
    } else {
      if (path != n2.path) {
        if (_debug) {
          print('## Equals: #3');
        }
        return false;
      }
      if (controller != n2.controller) {
        if (_debug) {
          print('## Equals: #14');
        }
        return false;
      }
    }
    return true;
  }

  @override
  Future<void> update(Node n2, {bool deepClone = true}) async {
    // clone all fields available in skeletoized state
    _controller = n2.controller;
    _skeleton.set(n2.skeleton);
    for (Field so in _skeletonOrdinals) {
      if ((n2 as NodeImpl)._ordinals[so] != null) {
        _ordinals[so] = n2._ordinals[so]!;
      }
    }
    if (!n2.skeleton) {
      // clone all ordinals
      _ordinals.clear();
      for (MapEntry<Field, String> e in (n2 as NodeImpl)._ordinals.entries) {
        _ordinals[e.key] = e.value;
      }

      //clone all values
      _values.clear();
      for (MapEntry<String, NodeValue> e in n2._values.entries) {
        _values[e.key] = e.value.deepClone();
      }

      // clone child nodes
      Map<String, Node> _newChildNodes = <String, Node>{};
      Map<String, Node> l = await n2.getChildren();
      for (String key in l.keys) {
        if (deepClone) {
          _newChildNodes[key] = await l[key]!.deepClone();
        } else {
          _newChildNodes[key] =
              NodeImpl.createSkeleton(l[key]!.path, n2.owner, controller);
        }
      }
      _childNodes.clear();
      _childNodes.addAll(_newChildNodes);
    }

    // make sure that lastModified is not altered by updates
    var lastModified = (n2 as NodeImpl)._ordinals[Field.lastModified];
    if (lastModified != null) {
      _ordinals[Field.lastModified] = lastModified;
    }
  }

  @override
  Future<Node> deepClone() async {
    return NodeImpl.fromNode(this);
  }

  @override
  Future<Node> shallowClone() async {
    Node ret = await NodeImpl.fromPath(path, owner);
    await ret.update(this, deepClone: false);
    return ret;
  }

  void touch() {}

  @override
  String toString() {
    var sb = StringBuffer();
    sb.write(path);
    sb.write('[');
    if (skeleton) {
      sb.write('{<skeletonized>}');
      sb.write(']{');
      sb.write('\n');
    } else {
      sb.write('owner=');
      sb.write(_ordinals[Field.owner]);
      sb.write(';vis=');
      sb.write(_ordinals[Field.visibility]);
      sb.write(']{');
      sb.write('\n');
    }
    var i = 0;
    if (skeleton) {
      sb.write('{<skeletonized>}');
    } else {
      for (var e in _values.entries) {
        if (i > 0) {
          sb.write(', ');
          sb.write('\n');
        }
        sb.write(e.value.toString('  '));
        i++;
      }
      sb.write('\n');
      sb.write('}');
    }
    if (_childNodes.isNotEmpty) {
      sb.write(' children{\n');
      int i = 0;
      for (Node element in _childNodes.values) {
        sb.write('  ${element.name}${i++ == 0 ? "" : ","}\n');
      }
      sb.write('}');
    }
    return sb.toString();
  }

  @override
  int get hashCode => toString().hashCode;

  @override
  bool operator ==(Object other) {
    if (other.runtimeType != NodeImpl) {
      return false;
    }
    return toString() == (other as NodeImpl).toString();
  }

  @override
  void toByteArrayStream(ByteSink out) {
    SerializerHelper.writeLong(out, serialversionUID);
    SerializerHelper.writeInt(out, skeleton ? 1 : 0);
    SerializerHelper.writeString(out, path);
    SerializerHelper.writeInt(out, _ordinals.length - 1);
    for (MapEntry<Field, String> e in _ordinals.entries) {
      if (e.key != Field.path) {
        SerializerHelper.writeString(out, e.key.toString());
        SerializerHelper.writeString(out, e.value);
      }
    }

    if (!skeleton) {
      SerializerHelper.writeInt(out, _values.length);
      for (MapEntry<String, NodeValue> e in _values.entries) {
        SerializerHelper.writeString(out, e.key);
        e.value.toByteArrayStream(out);
      }
    }

    SerializerHelper.writeInt(out, _childNodes.length);
    for (MapEntry<String, Node> e in _childNodes.entries) {
      SerializerHelper.writeString(out, e.key);
      e.value.toByteArrayStream(out);
    }

    SerializerHelper.writeLong(out, serialversionUID);
  }

  /// <p>Deserializes a NodeValue from a byteStream.</p>
  /// @param in the stream to be read
  /// @return the deserialized NodeValue
  /// @throws IOException if an exception happens deserializing the stream
  static Future<NodeImpl> fromByteArrayStream(ByteStream in_) async {
    if (await SerializerHelper.readLong(in_) != serialversionUID) {
      throw Exception('failed to parse NodeImpl (bad stream?)');
    }
    var skel = (await SerializerHelper.readInt(in_) == 1);
    var n = NodeImpl(await SerializerHelper.readString(in_) ?? '', 'NONE');
    if (skel) {
      n.controller = GenericController.getDefault();
    }
    int counter = await SerializerHelper.readInt(in_);
    for (var i = 0; i < counter; i++) {
      n._ordinals[
          Field.fromString(await SerializerHelper.readString(in_) ?? '') ??
              Field.owner] = await SerializerHelper.readString(in_) ?? '';
    }
    if (!skel) {
      counter = await SerializerHelper.readInt(in_);
      for (var i = 0; i < counter; i++) {
        n._values[await SerializerHelper.readString(in_) ?? ''] =
            await NodeValueImpl.fromByteArrayStream(in_);
      }
      counter = await SerializerHelper.readInt(in_);
      for (var i = 0; i < counter; i++) {
        n._childNodes[await SerializerHelper.readString(in_) ?? ''] =
            await NodeImpl.fromByteArrayStream(in_);
      }
    }
    if (await SerializerHelper.readLong(in_) != serialversionUID) {
      throw Exception('failed to parse NodeImpl (bad stream end?)');
    }
    return n;
  }
}
