library geiger_localstorage;

import 'dart:collection';

import 'package:geiger_localstorage/src/search_criteria.dart';
import 'package:geiger_localstorage/src/storage_listener.dart';
import 'package:uuid/uuid.dart';

import '../change_registrar.dart';
import '../event_type.dart';
import '../search_criteria.dart';
import '../storage_controller.dart';
import '../storage_exception.dart';
import 'data/node.dart';
import 'data/node_implementation.dart';
import 'data/node_value.dart';
import 'data/node_value_implementation.dart';
import 'storage_mapper.dart';

/// This Class Acts as an intermediate class to relay storageRequests to the StorageMapper.
///
/// This class contains the interaction API every actor (sensor, shields, UI,...)
/// calls methods from this controller to manipulate data.
class GenericController implements StorageController, ChangeRegistrar {
  static StorageController? defaultController;

  /// the path delimiter.
  static final String pathDelimiter = ':';

  /// the default owner of all newly created nodes.
  final String _owner;

  /// a map of all registered listeners for changes.
  final Map<SearchCriteria, StorageListener> listeners = HashMap();

  /// the storage mapper to be used.
  final StorageMapper mapper;

  int _inited = 0;

  /// <p>Construct the controller according to the database that is used.</p>
  ///
  /// @param owner  the owner to be used for all requests
  /// @param mapper the database mapper to use
  ///
  /// @throws StorageException In case of problems regarding the storage
  GenericController(this._owner, this.mapper) {
    mapper.setController(this);
    defaultController = this;
  }

  /// <p>Returns the latest created controller within the JVM.</p>
  ///
  /// @return a valid storage controller
  static StorageController? getDefault() {
    return defaultController;
  }

  Future<void> _initMapper() async {
    if (_inited==0) {
      _inited==1;
      final baseNodes = <String>[
        '',
        '${pathDelimiter}Devices',
        '${pathDelimiter}Users',
        '${pathDelimiter}Enterprise',
        '${pathDelimiter}Keys',
        '${pathDelimiter}Global',
        '${pathDelimiter}Local'
      ];
      for (var nodeName in baseNodes) {
        try {
          // for correct path generation creation of a Node is needed
          Node tmp = await NodeImpl.fromPath(nodeName, _owner);
          await mapper.get(tmp.path);
        } on StorageException {
          // node does not exist and therefore we create it
          Node tmp = await NodeImpl.fromPath(nodeName, _owner);
          tmp.owner = _owner;
          await mapper.add(tmp);
        }
      }

      // check if current user exists
      Node localNode = await mapper.get('${pathDelimiter}Local');
      NodeValue? uuid = await (localNode.getValue('currentUser'));
      if (uuid == null) {
        uuid = NodeValueImpl('currentUser', Uuid().v4());
        await localNode.addOrUpdateValue(uuid);
        localNode.owner = _owner;
        await mapper.update(localNode);
      }

      // create new default user
      uuid = NodeValueImpl('currentUser', Uuid().v4());
      await localNode.addOrUpdateValue(uuid);
      localNode.owner = _owner;
      await mapper.update(localNode);

      // check if current user node exists
      var userNodeName = '${pathDelimiter}Users$pathDelimiter' + uuid.value;
      try {
        await mapper.get(userNodeName);
      } on StorageException {
        Node n = await NodeImpl.fromPath(userNodeName, _owner);
        await mapper.add(n);
      }

      // check if current device exists
      localNode = await mapper.get('${pathDelimiter}Local');
      uuid = await localNode.getValue('currentDevice');
      if (uuid == null) {
        // create new default device
        uuid = NodeValueImpl('currentDevice', Uuid().v4());
        await localNode.addOrUpdateValue(uuid);
        localNode.owner = _owner;
        await mapper.update(localNode);
      }

      // check if current device node exists
      var deviceNodeName = '${pathDelimiter}Devices$pathDelimiter' + uuid.value;
      try {
        await mapper.get(deviceNodeName);
      } on StorageException {
        await mapper.add(await NodeImpl.fromPath(deviceNodeName, _owner));
      }
      _inited=2;
    }else if(_inited==1) {
      await Future.doWhile(() => _inited==2);
    }
  }

  @override
  Future<bool> addOrUpdate(Node? node) async {
    await _initMapper();
    if (node == null) {
      return false;
    } else if (node.skeleton) {
      return false;
    } else if (node.tombstone) {
      await delete(node.path);
      return false;
    } else {
      bool ret = false;
      try {
        await add(node);
        ret = true;
      } on StorageException {
        await update(node);
      }
      for (Node n2 in (await node.getChildren()).values) {
        ret |= await addOrUpdate(n2);
      }
      return ret;
    }
  }

  @override
  Future<Node> get(String path) async {
    await _initMapper();
    Node n = await getNodeOrTombstone(path);
    if (n.tombstone) {
      throw StorageException(
          'Node "$path" does not exist (but exited in the past; It is a tombstone now)');
    }
    List<String> l = <String>[];
    for (Node cn in (await n.getChildren()).values) {
      if (cn.tombstone) {
        l.add(cn.name);
      }
    }
    for (String name in l) {
      await n.removeChild(name);
    }
    return n;
  }

  @override
  Future<Node> getNodeOrTombstone(String path) async {
    return mapper.get(path);
  }

  @override
  Future<void> add(Node node) async {
    await _initMapper();
    // make sure that there is an owner set
    if ('' == node.owner) {
      node.owner = _owner;
    }

    // add object
    await mapper.add(node);

    // add child nodes
    for (Node n in (await node.getChildren()).values) {
      await mapper.add(n);
    }

    _checkListeners(EventType.create, null, node, null, null);
  }

  @override
  Future<void> update(Node node) async {
    await _initMapper();
    // make sure that there is an owner set
    if ('' == node.owner) {
      node.owner = _owner;
    }

    // get old node for update events
    Node oldNode = await mapper.get(node.path);

    // write node
    await mapper.update(node);

    _checkListeners(EventType.update, oldNode, node, null, null);

    // any child that is not a skeleton will be handled as new or changed
    for (Node child in (await node.getChildren()).values) {
      if (!child.skeleton) {
        try {
          await add(child);
        } on StorageException {
          // node already exists, therefore it was changed
          await update(child);
        }
      }
    }
  }

  @override
  Future<Node> delete(String path) async {
    await _initMapper();
    Node ret = await mapper.delete(path);
    _checkListeners(EventType.delete, ret, null, null, null);
    return ret;
  }

  @override
  Future<void> rename(String oldPath, String newPathOrName) async {
    await _initMapper();
    Node oldNode = await mapper.get(oldPath);
    String newPath = newPathOrName;
    if (!newPathOrName.startsWith(pathDelimiter)) {
      // create path from name
      newPath = oldNode.parentPath! + pathDelimiter + newPathOrName;
    }
    await mapper.rename(oldPath, newPath);
    Node newNode = await mapper.get(newPath);
    _checkListeners(EventType.rename, oldNode, newNode, null, null);
  }

  @override
  Future<NodeValue?> getValue(String path, String key) async {
    return await mapper.getValue(path, key);
  }

  @override
  Future<void> addValue(String nodeName, NodeValue newValue) async {
    await _initMapper();
    Node oldNode = await mapper.get(nodeName);
    NodeValue? oldValue = await oldNode.getValue(newValue.key);
    if (oldValue != null) {
      throw StorageException('value "' + newValue.key + '" is already set');
    }
    Node newNode = await oldNode.deepClone();
    await newNode.addValue(newValue);
    await mapper.update(newNode);
    _checkListeners(EventType.update, oldNode, newNode, null, newValue);
  }

  @override
  Future<void> updateValue(String nodeName, NodeValue newValue) async {
    await _initMapper();
    Node oldNode = await mapper.get(nodeName);
    NodeValue? oldValue = await oldNode.getValue(newValue.key);
    if (oldValue == null) {
      throw StorageException('value "' + newValue.key + '" does not yet exist');
    }
    Node newNode = await oldNode.deepClone();
    await newNode.removeValue(newValue.key);
    await newNode.addValue(newValue);
    await mapper.update(newNode);
    _checkListeners(EventType.update, oldNode, newNode, oldValue, newValue);
  }

  @override
  Future<NodeValue> deleteValue(String nodeName, String key) async {
    await _initMapper();
    Node oldNode = await mapper.get(nodeName);
    NodeValue? oldValue = await oldNode.getValue(key);
    if (oldValue == null) {
      throw StorageException('value "' + key + '" does not yet exist');
    }
    Node newNode = await oldNode.deepClone();
    await newNode.removeValue(key);
    await mapper.update(newNode);
    _checkListeners(EventType.update, oldNode, newNode, oldValue, null);
    return oldValue;
  }

  @override
  Future<List<Node>> search(SearchCriteria criteria) async {
    return await mapper.search(criteria);
  }

  @override
  Future<void> close() async {
    await mapper.close();
  }

  @override
  Future<void> flush() async {
    await _initMapper();
    await mapper.flush();
  }

  Future<void> _checkListeners(final EventType event, final Node? oldNode,
      final Node? newNode, NodeValue? oldValue, NodeValue? newValue) async {
    if (oldNode == null || oldNode != newNode) {
      for (MapEntry<SearchCriteria, StorageListener> e in listeners.entries) {
        try {
          if ((oldNode != null &&
                  newNode != null &&
                  (await e.key.evaluate(oldNode) ||
                      await e.key.evaluate(newNode))) ||
              (oldNode != null && await e.key.evaluate(oldNode)) ||
              (newNode != null && await e.key.evaluate(newNode))) {
            try {
              e.value.gotStorageChange(event, oldNode, newNode);
            } catch (e) {
              // TODO(mgwerder): raise sensible exception
              // FIXME do something sensible (should not happen anyway)
              rethrow;
            }
          }
        } on StorageException {
          // TODO(mgwerder): raise sensible exception
          // FIXME do something sensible (should not happen anyway)
        }
      }
      // }
    }
  }

  @override
  void registerChangeListener(
      StorageListener listener, SearchCriteria criteria) {
    // synchronized(listeners) {
    listeners[criteria] = listener;
    // }
  }

  @override
  List<SearchCriteria> deregisterChangeListener(StorageListener listener) {
    List<SearchCriteria> remove = <SearchCriteria>[];
    for (MapEntry<SearchCriteria, StorageListener> e in listeners.entries) {
      if (e.value == listener) {
        remove.add(e.key);
      }
    }
    if (remove.isEmpty) {
      throw StorageException('Listener not registered');
    }
    for (var c in remove) {
      listeners.remove(c);
    }
    return remove;
  }

  @override
  Future<void> zap() async {
    await mapper.zap();
    await _initMapper();
  }
}
