library geiger_localstorage;

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:geiger_localstorage/src/db/data/node_value_implementation.dart';
import 'package:intl/locale.dart';
import 'package:sqlite3/sqlite3.dart';

/// This class maps the DBinterface functions to a SQLite database
class SqliteMapper extends AbstractSqlMapper {
  static const int maxFieldSize = 1024;
  static final String initString = 'CREATE TABLE IF NOT EXISTS storage_node (\n'
          'path TEXT CHECK( LENGTH(path) <=' +
      maxFieldSize.toString() +
      ') PRIMARY KEY,\n'
          'owner TEXT CHECK( LENGTH(owner) <= 40),\n'
          'name TEXT CHECK( LENGTH(name) <= 40) NOT NULL,\n'
          "visibility TEXT CHECK(visibility IN ('black', 'red', 'amber', 'green', 'white')) NOT NULL DEFAULT 'red',\n"
          'children TEXT CHECK( LENGTH(children) <=' +
      maxFieldSize.toString() +
      ') NULL,\n'
          "tombstone TEXT CHECK( tombstone IN ('TRUE', 'FALSE')) DEFAULT 'FALSE'\n"
          ');\n';
  static final String initNodeValueString =
      'CREATE TABLE IF NOT EXISTS node_value (\n'
              'path TEXT CHECK( LENGTH(path) <=' +
          maxFieldSize.toString() +
          ') NOT NULL,\n'
              'key TEXT CHECK( LENGTH(key) <= 40) NOT NULL,\n'
              'value TEXT CHECK( LENGTH(value) <= 16384),\n'
              'type TEXT CHECK( LENGTH(type) <= 40),\n'
              'locale TEXT CHECK( LENGTH(locale) <= 10) NOT NULL,\n'
              'last_modified TEXT CHECK( LENGTH(last_modified) <= 20) NOT NULL,\n'
              'PRIMARY KEY (path, key),\n'
              'FOREIGN KEY (path) REFERENCES storage_node(path)\n'
              ');\n';
  static final String initTranslationString =
      'CREATE TABLE IF NOT EXISTS translation (\n'
              'path TEXT CHECK( LENGTH(path) <=' +
          maxFieldSize.toString() +
          ') NOT NULL,\n'
              'key TEXT CHECK( LENGTH(key) <= 40) NOT NULL,\n'
              "identifier TEXT CHECK(identifier IN ('VALUE', 'DESCRIPTION')) NOT NULL,\n"
              'locale TEXT CHECK( LENGTH(locale) <= 10) NOT NULL,\n'
              'translation TEXT CHECK( LENGTH(translation) <=' +
          maxFieldSize.toString() +
          ') NOT NULL,\n'
              'PRIMARY KEY (path, key, identifier, locale),\n'
              'FOREIGN KEY (path, key) REFERENCES node_value(path, key)\n'
              ');\n';

  late Database conn;
  StorageController? controller;

  SqliteMapper(String dbPath) {
    try {
      conn = sqlite3.open(dbPath);
      conn.execute('SELECT * FROM node_value LIMIT 1;');
    } on SqliteException {
      try {
        _initialize();
      } on StorageException {
        rethrow;
      }
    }
  }

  void _initialize() {
    try {
      conn.execute(initString);
      conn.execute(initNodeValueString);
      conn.execute(initTranslationString);
    } on SqliteException catch (e, ste) {
      try {
        String deleteString = 'DROP TABLE IF EXISTS translation;'
            'DROP TABLE IF EXISTS node_value;'
            'DROP TABLE IF EXISTS storage_node;';
        conn.execute(deleteString);
      } on SqliteException catch (e2, st) {
        throw StorageException(
            'Whoops... error while tearing down database', st, e2);
      }
      throw StorageException('Could not initialize database', ste, e);
    }
  }

  @override
  void setController(StorageController controller) {
    this.controller = controller;
  }

  @override
  Future<Node> get(String path) async {
    checkPath(path);
    getSanity(path);
    String sqlStatement =
        'SELECT path, owner, name, visibility, children, tombstone '
        'FROM storage_node WHERE path = ?';
    Node res;
    try {
      ResultSet resultSet = _requestResult(sqlStatement, [path]);
      if (resultSet.isEmpty) {
        throw StorageException('Node "$path" does not exist on sqlite backend');
      } else {
        String owner = (resultSet.first['owner'] ?? '<UNSET>') as String;
        if ('true' == resultSet.first['tombstone'].toLowerCase()) {
          res = await NodeImpl.fromPath(resultSet.first['path'] as String, owner,
              isTombstone: true);
          res.visibility=VisibilityExtension.valueOf(
                  resultSet.first['visibility']! as String) ??
              Visibility.red;
          return res;
        }
        res = await NodeImpl.fromPath(resultSet.first['path'] as String,owner);
        res.owner=owner;
        res.visibility=VisibilityExtension.valueOf(
                resultSet.first['visibility'] as String) ??
            Visibility.red;
        String children = resultSet.first['children'] as String;
        if (!('' == children)) {
          for (String childName in children.split(',')) {
            res.addChild(
                NodeImpl.createSkeleton(path + ':' + childName, owner, controller));
          }
        }
      }
    } on Exception catch (exception, s) {
      if (!(exception is StorageException || exception is SqliteException)) {
        rethrow;
      }
      throw StorageException(
          'Could not retrieve node "$path"', null, exception, s);
    }
    sqlStatement = ('SELECT path, key, value, type, locale, last_modified '
        'FROM node_value WHERE path = ?');
    ResultSet resultSetValues = _requestResult(sqlStatement, [path]);
    String key;
    try {
      for (Row valueRow in resultSetValues) {
        key = valueRow['key'] as String;
        NodeValue value = NodeValueImpl(
            key,
            valueRow['value'] as String,
            (valueRow['type'] ?? '') as String,
            '',
            int.parse(valueRow['last_modified'] as String));
        String sqlStatementTranslations =
            'SELECT path, key, identifier, locale, translation '
            'FROM translation WHERE (path = ? AND key = ?)';
        ResultSet resultSetTranslations =
            _requestResult(sqlStatementTranslations, [path, key]);
        try {
          for (Row translationRow in resultSetTranslations) {
            String identifier = translationRow['identifier'] as String;
            if ('VALUE' == identifier) {
              value.setValue(translationRow['translation'] as String,
                  Locale.parse(translationRow['locale'] as String));
            } else {
              if ('DESCRIPTION' == identifier) {
                value.setDescription(translationRow['translation'] as String,
                    Locale.parse(translationRow['locale'] as String));
              }
            }
          }
        } on SqliteException catch (e,st) {
          throw StorageException(
              'Could not retrieve description for node "' +
                  path +
                  '" and key "' +
                  key +
                  '"',null,
              e,st);
        }
        res.addValue(value);
      }
    } on SqliteException catch (e,st) {
      throw StorageException(
          'Could not retrieve values for node "' + path + '"', null,e,st);
    }
    return res;
  }

  @override
  Future<void> add(Node node) async {
    if (node.skeleton) {
      throw StorageException('Skeleton nodes cannot be added.');
    }
    checkPath(node.path);
    try {
      await get(node.path);
      throw StorageException('Node "${node.path}" already exists');
    } on Exception catch (e, st) {
      if ((e is! StorageException)) {
        throw StorageException('Node "${node.path}" already exists', null,e, st);
      }
    }
    if (node.parentPath != null && '' != node.parentPath) {
      Node parent = await get(node.parentPath!);
      await parent.addChild(node);
      await update(parent);
    }
    String sqlStatement = 'INSERT INTO '
        'storage_node(path, owner, name, visibility, children, tombstone) '
        'VALUES (?,?,?,?,?,?)';
    try {
      _requestResult(sqlStatement, [
        node.path,
         node.owner,
        node.name,
        (node.visibility).toValueString(),
        await node.getChildNodesCsv(),
        node.tombstone.toString().toUpperCase()
      ]);
    } on StorageException catch (e, s) {
      throw StorageException('Could not add new node "${node.path}"', null, e, s);
    }
    for (NodeValue nv in (await node.getValues()).values) {
      await addValue(node.path, nv);
    }
  }

  @override
  Future<void> update(Node node) async {
    if (node.skeleton) {
      throw StorageException('Skeleton nodes cannot be added, '
          'please materialize before adding');
    }
    checkPath(node.path);
    get(node.path);
    String sqlStatement =
        'UPDATE storage_node SET owner = ?, visibility = ?, children = ?, '
            'tombstone = ? WHERE path = ?';
    try {
      _requestResult(sqlStatement, [
        node.owner,
        (node.visibility).toValueString(),
        await node.getChildNodesCsv(),
        node.tombstone.toString().toUpperCase(),
        node.path
      ]);
      for (MapEntry<String,NodeValue> entry in (await node.getValues()).entries) {
        if (await getValue(node.path, entry.key) == null) {
          await addValue(node.path, entry.value);
        } else {
          await updateValue(node.path, entry.value);
        }
      }
    } on StorageException catch (e,st) {
      throw StorageException('Could not update node', null,e,st);
    }
  }

  @override
  Future<Node> delete(String path) async {
    var oldNode = await get(path);
    if ((await oldNode.getChildren()).isNotEmpty ) {
      throw StorageException(
          'Node does have ${(await oldNode.getChildren()).length} children... cannot remove ' +
              path);
    }
    for (NodeValue nv in (await oldNode.getValues()).values) {
      await deleteValue(path, nv.key);
    }
    Node deletedNode = await NodeImpl.fromPath(path, oldNode.owner, isTombstone: true);
    deletedNode.visibility=oldNode.visibility;
    await update(deletedNode);
    if ('' != oldNode.parentPath) {
      Node parentNode = await get(oldNode.parentPath!);
      await parentNode.removeChild(oldNode.name);
      await update(parentNode);
    }
    return oldNode;
  }

  @override
  Future<NodeValue?> getValue(String path, String key) async {
    if (('' == path) || ('' == key)) {
      throw NullThrownError();
    }
    checkPath(path);
    Node ret = await get(path);
    if (ret.tombstone) {
      throw StorageException(
          'unable to get value... Node "$path" does not exists. Node is a tombstone.');
    }
    String sqlSelectStatement = ('SELECT path,key,value,type,locale,last_modified '
        'FROM node_value WHERE (path = ? and key = ?)');
    NodeValue? value;
    try {
      ResultSet resultSet = _requestResult(sqlSelectStatement, [path, key]);
      if (resultSet.isEmpty) {
        return value;
      }
      value = NodeValueImpl(
          resultSet.first['key'] as String,
          resultSet.first['value'] as String,
          (resultSet.first['type'] ?? '') as String,
          '',
          int.parse(resultSet.first['last_modified'] as String));
    } catch (e) {
      if (!(e is StorageException && e is SqliteException)) rethrow;
      throw StorageException(
          'Something went wrong while trying to retrieve the value ' +
              key +
              ' from ' +
              path,
          null,e);
    }
    String sqlStatementTranslations =
        ('SELECT path,key,identifier,locale,translation '
            'FROM translation WHERE (path = ? AND key = ?)');
    try {
      ResultSet rsTranslations = _requestResult(sqlStatementTranslations, [path, key]);
      for (Row rsTranslation in rsTranslations) {
        String identifier = rsTranslation['identifier'] as String;
        if ('VALUE' == identifier) {
          value.setValue(rsTranslation['translation'] as String,
              Locale.parse(rsTranslation['locale'] as String));
        } else {
          if ('DESCRIPTION' == identifier) {
            value.setDescription(rsTranslation['translation'] as String,
                Locale.parse(rsTranslation['locale'] as String));
          }
        }
      }
    } on SqliteException catch (e) {
      throw StorageException(
          'Could not retrieve description for node "' +
              path +
              '" and key "' +
              key +
              '"',null,
          e);
    }
    return value;
  }

  Future<void> addValue(String path, NodeValue value) async {
    checkPath(path);
    await get(path);
    if (await getValue(path, value.key) != null) {
      throw StorageException('Value "${value.key}" already exists');
    }
    String sqlStatement =
        ('INSERT INTO node_value (path, key, value, type, locale, '
            'last_modified) VALUES (?,?,?,?,?,?)');
    try {
      _requestResult(sqlStatement, [
        path,
        value.key,
        value.value,
        value.type,
        NodeValueImpl.defaultLocale.toLanguageTag(),
        value.lastModified.toString()
      ]);
    } on StorageException catch (e) {
      throw StorageException('Could not create value "${value.key}"', null,e);
    }
    String sqlStatementTrl =
        ('INSERT INTO translation (path, key, identifier, locale, translation)'
            ' VALUES (?,?,?,?,?)');
    Map<Locale,String> valueMap = value.allValueTranslations;
    for (MapEntry<Locale,String> entry in valueMap.entries) {
      try {
        _requestResult(sqlStatementTrl,
            [path, value.key, 'VALUE', entry.key.toLanguageTag(), entry.value]);
      } on StorageException catch (e) {
        throw StorageException(
            'Could not create translation "' +
                entry.key.toLanguageTag() +
                '" for value "' +
                value.key +
                '"'
            ,null, e);
      }
    }
    Map<Locale,String> descriptionMap = value.allDescriptionTranslations;
    for (MapEntry<Locale,String> entry in descriptionMap.entries) {
      try {
        _requestResult(sqlStatementTrl, [
          path,
          value.key,
          'DESCRIPTION',
          entry.key.toLanguageTag(),
          entry.value
        ]);
      } on StorageException catch (e) {
        throw StorageException(
            'Could not create translation "' +
                entry.key.toLanguageTag() +
                '" for description in value "' +
                value.key +
                '"',
            null,e);
      }
    }
  }

  Future<void> updateValue(String path, NodeValue value) async {
    checkPath(path);
    if ('' == path) {
      throw NullThrownError();
    }
    if (await getValue(path, value.key) == null) {
      throw StorageException(('Key "' + value.key) + '" does not exist');
    }
    deleteValue(path, value.key);
    addValue(path, value);
  }

  Future<NodeValue> deleteValue(String path, String key) async {
    if ('' == path) {
      throw NullThrownError();
    }
    get(path);
    NodeValue? value = await getValue(path, key);
    if (value == null) {
      throw StorageException(('Key "' + key) + '" does not exist');
    }
    String sqlDeleteStatementTranslations =
        'DELETE FROM translation WHERE (path = ? AND key = ?)';
    try {
      _requestResult(sqlDeleteStatementTranslations, [path, key]);
    } on StorageException catch (e) {
      throw StorageException(
          'Could not delete translations for key "' +
              key +
              '" in node "' +
              path +
              '"',
          null,e);
    }
    String sqlDeleteStatement =
        'DELETE FROM node_value WHERE (path = ? AND key = ?)';
    try {
      _requestResult(sqlDeleteStatement, [path, key]);
    } on StorageException catch (e) {
      throw StorageException(
          'Could not delete value for key "' + key + '" in node "' + path + '"',
          null,e);
    }
    return value;
  }

  @override
  Future<void> rename(String oldPath, String newPathOrName) async {
    checkPath(oldPath);
    checkPath(newPathOrName);
    Node oldNode = await get(oldPath);
    Node newNode;
    if (oldNode.tombstone) {
      newNode = await NodeImpl.fromPath(newPathOrName, oldNode.owner, isTombstone: true);
    } else {
      String? path = NodeImpl.getNameFromPath(newPathOrName)==newPathOrName? NodeImpl.getParentFromPath(oldPath): NodeImpl.getParentFromPath(newPathOrName);
      newNode = NodeImpl(NodeImpl.getNameFromPath(newPathOrName)!, oldNode.owner, path,oldNode.visibility);
    }
    for (NodeValue nv in (await oldNode.getValues()).values) {
      newNode.addValue(nv);
    }
    add(newNode);
    for (Node n in (await oldNode.getChildren()).values) {
      rename(
          n.path, newNode.path + GenericController.pathDelimiter + n.name);
    }
    delete(oldPath);
  }

  @override
  Future<List<Node>> search(SearchCriteria criteria) async {
    String sqlNodeSearch =
        'SELECT path, owner, name, visibility, children FROM storage_node '
        'WHERE (path = ? and owner = ? and name = ? and visibility = ?)';
    String sqlValueSearch = 'SELECT path,key,value,type,locale,last_modified '
        'FROM node_value WHERE (path = ? and key = ? and value = ? and type = ? '
        'and last_modified = ?)';
    Map<String, Node> nodes = {};
    Map<String, NodeValue?> values = {};
    try {
      ResultSet resultSetNodes = _requestResult(sqlNodeSearch, [
        ('' == criteria.path) ? '*' : criteria.path,
        ('' == criteria.owner) ? '*' : criteria.owner,
        ('' == criteria.name) ? '*' : criteria.name,
        ('' == criteria.get(Field.visibility))
            ? '*'
            : criteria.get(Field.visibility)
      ]);
      for (Row node in resultSetNodes) {
        String path = node['path'] as String;
        nodes[path] = await get(path);
      }
      ResultSet resultSetValues = _requestResult(sqlValueSearch, [
        ('' == criteria.path) ? '*' : criteria.path,
        ('' == criteria.nodeValueKey) ? '*' : criteria.nodeValueKey,
        ('' == criteria.nodeValueValue) ? '*' : criteria.nodeValueValue,
        ('' == criteria.nodeValueType) ? '*' : criteria.nodeValueType,
        ('' == criteria.nodeValueLastModified)
            ? '*'
            : criteria.nodeValueLastModified
      ]);
      for (Row value in resultSetValues) {
        String path = value['path'] as String;
        values[path] = await getValue(path, value['key'] as String);
      }
    } on SqliteException catch (_, st) {
      throw StorageException(
          'Something went wrong while trying to execute the search', null,_,st);
    }
    for (String path in values.keys) {
      if (!nodes.containsKey(path)) {
        nodes[path] = await get(path);
      }
      nodes[path]!.addValue(values[path]!);
    }
    return nodes.values.toList();
  }

  @override
  Future<void> close() async {
    try {
      conn.dispose();
    } on SqliteException catch (e) {
      throw Exception('OOPS... unexpected exception' + e.toString());
    }
  }

  @override
  Future<void> flush() async {
    // TODO(mgwerder): implement flush
    /// try {
    //conn.commit();
    /// } catch (SQLException e) {
    /// throw new RuntimeException("Could not flush database", e);
    /// }
  }

  @override
  Future<void> zap() async {
    String sqlStatement1 = 'DELETE FROM storage_node';
    String sqlStatement2 = 'DELETE FROM node_value';
    String sqlStatement3 = 'DELETE FROM translation';
    try {
      conn.execute(sqlStatement3);
      conn.execute(sqlStatement2);
      conn.execute(sqlStatement1);
    } on SqliteException catch (e) {
      throw StorageException(
          'Something went wrong while trying to truncate the database', null, e);
    }
  }

  /// Executes the given sqlStatement as a prepared statement using the optional args in given order.
  ///
  /// @param sqlStatement the statement to execute
  /// @param args the arguments to replace
  /// @return the Resultset containing all the results
  ResultSet _requestResult(String sqlStatement, List<Object?> args) {
    try {
      return conn.select(sqlStatement, args);
    } on SqliteException catch (e, st) {
      throw StorageException(
          'Could not execute query: "$sqlStatement" with args: ${args.toString()} [${e.toString()}]',null,
          e,
          st);
    }
  }
}
