library geiger_localstorage;

import 'package:geiger_localstorage/geiger_localstorage.dart';

/// <p>Defines the type of comparator to be used when accessing an ordinal.</p>
enum ComparatorType { string, datetime, boolean }

/// <p>An object that can hold all possible search criteria.</p>
///
/// <p>Each criteria can either be set or left blank the search will match all
/// nonempty criteria.
/// </p>
class SearchCriteria with Serializer, Comparable<SearchCriteria> {
  static const int serialversionUID = 87128319541;

  final Map<Field, String> values = {};

  /// <p>Create a [SearchCriteria].</p>
  ///
  /// @param path the path to search for
  /// @param key the key to search for
  /// @param value the value to search for
  SearchCriteria([String? searchPath, String? key, String? value]) {
    if (searchPath == null && key == null) {
      return;
    }
    path = searchPath??':';
    nodeValueKey = key;
    if (value != null && value != '%') {
      nodeValueValue = value;
    }
  }

  String? get owner {
    return values[Field.owner];
  }

  set owner(String? nodeOwner) {
    if (nodeOwner == null) {
      values.remove(Field.owner);
    } else {
      values[Field.owner] = nodeOwner;
    }
  }

  String get name {
    return values[Field.name] ?? 'OOOPS#1';
  }

  set name(String nodeName) {
    values[Field.name] = nodeName;
  }

  String get path {
    return values[Field.path] ?? 'OOOPS#2';
  }

  set path(String nodePath) {
    values[Field.path] = nodePath;
  }

  String? get nodeValueKey {
    return values[Field.key];
  }

  set nodeValueKey(String? nodeValueKey) {
    if (nodeValueKey == null) {
      values.remove(Field.key);
    } else {
      values[Field.key] = nodeValueKey;
    }
  }

  String? get nodeValueValue {
    return values[Field.value];
  }

  set nodeValueValue(String? nodeValue) {
    if (nodeValue == null) {
      values.remove(Field.value);
    } else {
      values[Field.value] = nodeValue;
    }
  }

  String? get nodeValueType {
    return values[Field.type];
  }

  set nodeValueType(String? nodeValueType) {
    if (nodeValueType == null) {
      values.remove(Field.type);
    } else {
      values[Field.type] = nodeValueType;
    }
  }

  String? get(Field f) {
    return values[f];
  }

  void set(Field f, String value) {
    values[f] = value;
  }

  String? get nodeValueLastModified {
    return values[Field.lastModified];
  }

  set nodeValueLastModified(String? nodeValueLastModified) {
    if (nodeValueLastModified == null) {
      values.remove(Field.lastModified);
    } else {
      values[Field.lastModified] = nodeValueLastModified;
    }
  }

  /// Evaluates a provided node against this criteria
  ///
  /// The node is evaluated as follows. An node is equivalent if and only if...
  ///
  /// 1. the node is the same path or a sub path of the search criteria path [path]
  /// 2. if an owner regexp is set -> it must match the nodes owner
  /// 3. if a visibility regexp is set -> it must match the visibility
  /// 4. if a [nodeValueKey] and a [nodeValueValue] is set ->
  /// 4a. The respective key must exist (exact! no regexp) and [nodeValueType] and [nodeValueValue] must match the respective regex.
  /// 5. if no [nodeValueKey] and a [nodeValueValue] is set ->
  /// 5a. Any key of the node must match [nodeValueType] and [nodeValueValue] the respective regex.
  ///
  /// @param node the node to be evaluated
  /// @return true if the node matches the criteria
  ///
  /// @throws StorageException if the storage backend encounters a problem
  Future<bool> evaluate(Node node) async {
    // check for equal path start (must be a node or subnode of path
    if (!(node.path).startsWith(path)) {
      return false;
    }
    // check owner
    if ((values[Field.owner] != null) &&
        (!regexEvalString(values[Field.owner] ?? '', node.owner))) {
      return false;
    }
    //check visibility
    String? visibility = values[Field.visibility];
    if ((visibility != null) &&
        (!regexEvalString(visibility, (node.visibility).toString()))) {
      return false;
    }
    // check if any of the node values matches
    Map<String,NodeValue> nodeValues = await node.getValues();
    if ((values[Field.key] == null) &&
        ((values[Field.value] != null) || (values[Field.type] != null))) {
      for (MapEntry<String,NodeValue> e in nodeValues.entries) {
        String? type = values[Field.type];
        bool r3 = type == null ||
            !regexEvalString(type, e.value.type ?? 'NoValidType');
        String? value = values[Field.value];
        bool r2 = value == null || !regexEvalString(value, e.value.value);
        if (r2 && r3) {
          return true;
        }
      }
      return false;
    } else {
      if (values[Field.key] != null) {
         NodeValue? nv = nodeValues[get(Field.key)];
        if(nv==null)  {
          return false;
        }
        if (!regexEvalString(values[Field.type]!, nv.type ?? 'NoValidType')) {
          return false;
        }
        if (!regexEvalString(values[Field.value]!, nv.value)) {
          return false;
        }
      }
    }
    return true;
  }

  bool regexEvalString(String regex, String value) {
    return RegExp(regex).hasMatch(value);
  }

  @override
  void toByteArrayStream(ByteSink out) {
    SerializerHelper.writeLong(out, serialversionUID);
    SerializerHelper.writeInt(out, values.length);
    for (MapEntry<Field, String> e in values.entries) {
      SerializerHelper.writeString(out, e.key.toString());
      SerializerHelper.writeString(out, e.value);
    }
    SerializerHelper.writeLong(out, serialversionUID);
  }

  /// <p>Static deserializer.</p>
  ///
  /// <p>creates  a search criteria from a ByteArrayStream</p>
  /// @param in The input byte stream to be used
  /// @return the object parsed from the input stream by the respective class
  /// @throws IOException if not overridden or reached unexpectedly the end of stream
  static Future<SearchCriteria> fromByteArrayStream(ByteStream in_) async {
    if (await SerializerHelper.readLong(in_) != serialversionUID) {
      throw StorageException( 'failed to parse StorageException (bad stream?)',null);
    }
    SearchCriteria s = SearchCriteria();
    int size = await SerializerHelper.readInt(in_);
    for (int i = 0; i < size; i++) {
      s.values[Field.fromString((await SerializerHelper.readString(in_))!)!]=(await SerializerHelper.readString(in_))!;
    }
    if (await SerializerHelper.readLong(in_) != serialversionUID) {
      throw StorageException(
          'failed to parse StorageException (bad stream end?)');
    }
    return s;
  }

  @override
  String toString() {
    String sb = '{\n';
    Set<String> tmp = {};
    for (Field f in values.keys) {
      tmp.add(f.toString());
    }
    for (String f in tmp) {
      sb+='  ${f.toString()}=${values[Field.fromString(f)]}\n';
    }
    sb+='}\n';
    return sb;
  }

  @override
  int compareTo(SearchCriteria other) {
    return toString().compareTo(other.toString());
  }

  bool equals(Object o) {
    if (o is! SearchCriteria) {
      return false;
    }
    return toString() == o.toString();
  }
}
