import 'dart:io';

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:geiger_localstorage/src/db/data/node_implementation.dart';
import 'package:geiger_localstorage/src/db/data/node_value_implementation.dart';
import 'package:geiger_localstorage/src/db/mapper/dummy_mapper.dart';
import 'package:geiger_localstorage/src/db/mapper/sqlite_mapper.dart';
import 'package:geiger_localstorage/src/db/storage_mapper.dart';
import 'package:geiger_localstorage/src/storage_exception.dart';
import 'package:intl/locale.dart';
import 'package:test/test.dart';

class StorageMapperTest {}

void _testPrep(List<StorageMapper> mapperList) {
  for (final mapper in mapperList) {
    mapper.zap();
  }
}

void _testTearDown(List<StorageMapper> mapperList) {
  for (final mapper in mapperList) {
    mapper.flush();
  }
}

void main() {
  late List<StorageMapper> mapperList;

  setUp(() {
    mapperList = [DummyMapper('testowner'), SqliteMapper('./testdb.sqlite')];
    _testPrep(mapperList);
  });

  test('Adding node', () async {
    for (final mapper in mapperList) {
      print('## Testing mapper $mapper in UNKNOWN');
      Node node = NodeImpl('testNode1', 'testowner', '');
      node.owner = 'myOwner';
      print("### adding new node");
      await mapper.add(node);

      // make sure that lastupdated and owner is not updated upon add
      print("### retrieving previously written node");
      Node node2 = await mapper.get(':testNode1');
      print("### testing node content");
      expect(node2.equals(node), true,
          reason: 'Last updated was unexpectedly updated');
    }
  });

  test('add root node null', () async {
    for (final mapper in mapperList) {
      print('## Testing mapper ' + mapper.toString() + ' in UNKNOWN');
      Node node = NodeImpl('testNode1', 'testowner');
      node.owner = 'myOwner';

      await mapper.add(node);

      // make sure that lastupdated and owner is not updated upon add
      Node node2 = await mapper.get(':testNode1');
      expect(node2.equals(node), true,
          reason: 'Last updated was unexpectedly updated');
    }
  });

  test('add root node illegal name', () {
    for (final StorageMapper mapper in mapperList) {
      for (String name in [':name', 'na:me', 'name:']) {
        print('## Testing mapper ' + mapper.toString() + ' in UNKNOWN');
        Node node = NodeImpl(name, 'testowner', '');
        expect(() async => await mapper.add(node),
            throwsA(TypeMatcher<StorageException>()));
        // make sure that the element was not added
        expect(() async => await mapper.get(name),
            throwsA(TypeMatcher<StorageException>()));
      }
    }
  });
  test('add root node illegal parent', () async {
    for (final StorageMapper mapper in mapperList) {
      await mapper.add(NodeImpl('na', ''));
      for (var parent in [':name:', 'na:me', 'name:', '::', ':na::']) {
        print('## Testing mapper ' + mapper.toString() + ' in INVALID');
        Node node = NodeImpl('name', 'testowner', parent);
        expect(() async => await mapper.add(node),
            throwsA(TypeMatcher<StorageException>()));
        // make sure that the element was not added
        expect(() async => await mapper.get(parent + ':name'),
            throwsA(TypeMatcher<StorageException>()));
      }
    }
  });
  test('add duplicate node', () async {
    for (final StorageMapper mapper in mapperList) {
      print('## Testing mapper ' + mapper.toString() + ' in INVALID');
      Node node = NodeImpl('testNode1', '');
      await mapper.add(node);
      expect(() async => await mapper.add(node),
          throwsA(TypeMatcher<StorageException>()));
    }
  });
  test('add parent-less node', () {
    for (final StorageMapper mapper in mapperList) {
      print('## Testing mapper ' + mapper.toString() + ' in INVALID');
      Node node = NodeImpl('testNode1', 'testOwner', ':notAValidParent');
      expect(() async => await mapper.add(node),
          throwsA(TypeMatcher<StorageException>()));
    }
  });
  test('node getter tests', () async {
    for (final mapper in mapperList) {
      print('## Testing mapper ' + mapper.toString() + ' in INVALID');
      Node node = NodeImpl('testNode1', 'testOwner', '');
      Node childNode = NodeImpl('testNode1a', 'testOwner', ':testNode1');
      NodeValue nv = NodeValueImpl('key', 'value', 'type', 'description', 1);
      await node.addChild(childNode);
      await node.addValue(nv);

      // write data
      Node node2 = NodeImpl('testNode2', 'testOwner', '');
      await mapper.add(node);
      await mapper.add(node2);
      await mapper.add(childNode);

      // get data
      Node storedNode = await mapper.get(':testNode1');
      Node storedChildNode = await mapper.get(':testNode1:testNode1a');
      Node storedNode2 = await mapper.get(':testNode2');
      expect(storedNode.skeleton, false,
          reason: 'Node :testNode1 may not be a skeleton if fetched');
      expect(storedChildNode.skeleton, false,
          reason: 'Node :testNode1 may not be a skeleton if fetched');
      expect(storedNode2.skeleton, false,
          reason: 'Node :testNode1 may not be a skeleton if fetched');

      // compare data
      expect(node, storedNode,
          reason:
              'comparing parent node failed (expected: $node; got: $storedNode');
      expect(childNode.equals(storedChildNode), isTrue,
          reason: 'comparing child node');
      expect(node2.equals(storedNode2), isTrue,
          reason: 'comparing parent node2 ');
      expect((await storedNode.getChildren()).length, 1,
          reason: 'checking for child node count');
    }
  });
  test('persistence of locales', () async {
    for (final StorageMapper mapper in mapperList) {
      print('## Testing mapper ' + mapper.toString() + ' in INVALID');
      var nv = NodeValueImpl('key', 'value', 'type', 'description', 1);
      nv.setDescription('Deutsche Beschreibung', Locale.parse('de-de'));
      nv.setDescription('Schwiizertüütschi Beschriibig', Locale.parse('de-ch'));
      nv.setValue('Wert', Locale.parse('de-de'));
      nv.setValue('Au ä Wert', Locale.parse('de-ch'));
      Node node = NodeImpl('testNode1', 'testOwner', '');
      Node childNode = NodeImpl('testNode1a', 'testOwner', ':testNode1');
      await node.addChild(childNode);
      await node.addValue(nv);
      print('## Stored ' + node.toString());

      // write data
      Node node2 = NodeImpl('testNode2', 'testOwner', '');
      await mapper.add(node);
      await mapper.add(node2);
      await mapper.add(childNode);

      // get data
      Node storedNode = await mapper.get(':testNode1');
      Node storedChildNode = await mapper.get(':testNode1:testNode1a');
      Node storedNode2 = await mapper.get(':testNode2');
      expect(storedNode.skeleton, false,
          reason: 'Node :testNode1 may not be a skeleton if fetched');
      expect(storedChildNode.skeleton, false,
          reason:
              'Node :testNode1:testNode1a may not be a skeleton if fetched');
      expect(storedNode2.skeleton, false,
          reason: 'Node :testNode2 may not be a skeleton if fetched');
      print('## got    ' + storedNode.toString());

      // compare data
      expect(node,storedNode,
          reason:
              'comparing parent node (expected: ${node.toString()}(${node.toString().hashCode}); got: ${storedNode.toString()}(${storedNode.toString().hashCode})');
      expect( childNode.equals(storedChildNode), true,
          reason: 'comparing child node');
      expect(node2.equals(storedNode2), true,
          reason: 'comparing parent node2');
      expect((await storedNode.getChildren()).length, 1,
          reason: 'checking for child node count');
    }
  });
  tearDown(() async {
    _testTearDown(mapperList);

    // tear down all test data
    for (final StorageMapper mapper in mapperList) {
      await mapper.close();
    }
    File('./testdb.sqlite').deleteSync();
  });
}
